#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xft/Xft.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

char *getKeyString(KeySym keysym);
char *HexAddress(int address_num, int size);

int main(int argc, char **argv) {
    // Program Name
    char           *progname;
    
    // Display
    Display        *display;
    char           *hostname;
    Visual         *visual;
    Colormap        colormap;
    XEvent          event;
    KeySym          key;
    int             screen_nr;
    unsigned int    display_width;
    unsigned int    display_height;
    
    // Window
    Window          win;
    Window          root;
    
    XWindowChanges  win_attrib;
    
    unsigned long   color_black;
    unsigned long   color_white;
    char           *str_name1 = "Alpha";
    char           *str_name2 = "Beta";
    char           *key_string;
    int             msg_x;
    int             msg_y;
    int             error_code;
    
    // Hints
    XSizeHints     *size_hints;
    XWMHints       *wm_hints;
    XClassHint     *class_hint;
    XTextProperty  *win_name;
    XTextProperty  *ico_name;
    
    // X Free Type
    XftDraw        *xftdraw;
    XftFont        *xftfont;
    XftColor        color_fg;
    XGlyphInfo      xftinfo;
    
    // 
    progname = argv[0];
    if (argc > 1) {
        hostname = argv[1];
    } else {
        hostname = NULL;
    }
    
    // Alloc Size Hints
    if ((size_hints = XAllocSizeHints()) == False) {
        fprintf(stderr, "%s: failure allocating memory\n", progname);
        exit(EXIT_FAILURE);
    }
    
    // Alloc WM Hints
    if ((wm_hints = XAllocWMHints()) == False) {
        fprintf(stderr, "%s: failure allocating memory\n", progname);
        exit(EXIT_FAILURE);
    }
    
    // Alloc Class Hint
    if ((class_hint = XAllocClassHint()) == False) {
        fprintf(stderr, "%s: failure allocating memory\n", progname);
        exit(EXIT_FAILURE);
    }
    
    if ((display = XOpenDisplay(hostname)) == NULL) {
        fprintf(stderr, "%s: cannot connect to X server %s\n", progname, XDisplayName(hostname));
        exit(EXIT_FAILURE);
    }
    
    // Default
    screen_nr      = DefaultScreen(display);
    display_width  = DisplayWidth(display, screen_nr);
    display_height = DisplayHeight(display, screen_nr);
    root           = RootWindow(display, screen_nr);
    color_black    = BlackPixel(display, screen_nr);
    color_white    = WhitePixel(display, screen_nr);
    visual         = DefaultVisual(display, screen_nr);
    colormap       = DefaultColormap(display, screen_nr);
    
    // Create Window
    win_attrib.width      = 800;
    win_attrib.height     = 600;
    win_attrib.x          = (display_width  - win_attrib.width)  / 2;
    win_attrib.y          = (display_height - win_attrib.height) / 2;
    
    win = XCreateSimpleWindow(display, root,
                              win_attrib.x, win_attrib.y,
                              win_attrib.width, win_attrib.height, 0,
                              color_black, color_white);
    
    // Create Xft Drawing
    xftdraw = XftDrawCreate(display, (Drawable) win, visual, colormap);
    
    // Load Font
    xftfont = XftFontOpen(display, screen_nr,
                          XFT_FAMILY, XftTypeString,  "Arial",
                          XFT_SIZE,   XftTypeInteger, 25, 0);
    
    // Set Color
    color_fg.color.red   = 0x00ff;
    color_fg.color.green = 0x00ff;
    color_fg.color.blue  = 0x00ff;
    color_fg.color.alpha = 0x00ff00;
    color_fg.pixel       = 0x00ff00;
    
    // Set Size Hints
    size_hints->flags  = USPosition | USSize;
    size_hints->width  = win_attrib.width;
    size_hints->height = win_attrib.height;
    size_hints->x      = win_attrib.x;
    size_hints->y      = win_attrib.y;
    
    // Set WM Hints
    wm_hints->flags         = InputHint | StateHint;
    wm_hints->input         = True;
    wm_hints->initial_state = NormalState;
    
    // Set Class Hint (not used)
    class_hint->res_name  = str_name1;
    class_hint->res_class = str_name2;
    
    // Window Name
    win_name = malloc(sizeof(XTextProperty));
    if (XStringListToTextProperty(&str_name1, 1, win_name) == 0) {
        fprintf(stderr, "%s: structure allocation for windowName failed.\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    
    // Icon Name
    ico_name = malloc(sizeof(XTextProperty));
    if (XStringListToTextProperty(&str_name2, 1, ico_name) == 0) {
        fprintf(stderr, "%s: structure allocation for iconName failed.\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    
    // Set hints for window manager before mapping window
    XSetWMProperties(display, win, win_name, ico_name, NULL, 0, size_hints, wm_hints, class_hint);
    
    // Select Event Types
    XSelectInput(display, win, KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask |
                 ExposureMask | StructureNotifyMask | FocusChangeMask);
    
    // Display Window
    XMapWindow(display, win);
    
    // Event Loop
    while (True) {
        XNextEvent(display, &event);
        
        switch (event.type) {
            case KeyPress:
                error_code = XLookupString(&event.xkey, NULL, 0, &key, NULL);
                key_string = getKeyString(key);
                
                // Get Xft Font Info
                XftTextExtents8(display, xftfont, key_string, strlen(key_string), &xftinfo);
                
                // Output message centrally in window
                msg_x = (win_attrib.width  - xftinfo.width)  / 2;
                msg_y = (win_attrib.height - xftinfo.height) / 2;
                
                XftDrawStringUtf8(xftdraw, &color_fg, xftfont, msg_x, msg_y,
                                 (unsigned char *) key_string,
                                 strlen(key_string));
                if (key == 'q') {
                    XFree(win_name);
                    XFree(ico_name);
                    XDestroyWindow(display, win);
                    XCloseDisplay(display);
                    printf("Exiting...\n\n");
                    exit(EXIT_SUCCESS);
                }
                
                break;
                
            case MappingNotify:
                XRefreshKeyboardMapping(&event.xmapping);
                break;
                
            default:
        }
    }
}

char *getKeyString(KeySym keysym) {
    if (keysym >= 0x020 && keysym <= 0x0ff) {
        char *s = malloc(2 * sizeof(char));
        s[0] = (char) keysym;
        s[1] = '\0';
        return s;
    }
    
    switch (keysym) {
        case XK_BackSpace:   return "Backspace";
        case XK_Tab:         return "Tab";
        case XK_Linefeed:    return "Linefeed";
        case XK_Clear:       return "Clear";
        case XK_Return:      return "Enter";
        case XK_Pause:       return "Pause";
        case XK_Scroll_Lock: return "Scroll Lock";
        case XK_Sys_Req:     return "System Requirement";
        case XK_Escape:      return "Escape";
        case XK_Delete:      return "Delete";
    }
    return "Unknow";
}
